import json
import sys

import anyio
import click
import mcp.types as types
from dap_types import LaunchRequestArguments, SourceBreakpoint
from mcp.server.lowlevel import Server
from pathlib import Path
from pydantic import TypeAdapter
from typing import Optional, TextIO, Literal

from dap_mcp.config import DebuggerSpecificConfig
from dap_mcp.debugger import Debugger, FunctionCallError
from dap_mcp.factory import DAPClientSingletonFactory
from dap_mcp.render import render_xml, try_render

import logging

logger = logging.getLogger(__name__)


@click.command()
@click.option("--port", default=8000, help="Port to listen on for SSE")
@click.option(
    "--transport",
    type=click.Choice(["stdio", "sse"]),
    default="stdio",
    help="Transport type",
)
@click.option(
    "--verbose", "-v", is_flag=True, help="Enable verbose logging", default=False
)
@click.option(
    "--config",
    "-c",
    "config_file",
    help="Path to the configuration file",
    required=True,
    type=click.File("r"),
)
def main(
    port: int,
    transport: str,
    verbose: bool,
    config_file: TextIO,
) -> int:
    if verbose:
        logging.basicConfig(level=logging.DEBUG, stream=sys.stderr)
    debug_config_type_adapter: TypeAdapter[DebuggerSpecificConfig] = TypeAdapter(
        DebuggerSpecificConfig
    )
    try:
        json_config = json.load(config_file)
    except json.JSONDecodeError as e:
        logger.error(f"Error parsing JSON configuration: {e}")
        return 1
    config = debug_config_type_adapter.validate_python(json_config)
    app: Server[object] = Server("dap_mcp")
    dap_factory = DAPClientSingletonFactory(config.debuggerPath, config.debuggerArgs)
    launch_arguments = LaunchRequestArguments(
        noDebug=False,
        **config.model_dump(
            exclude_none=True,
            exclude={"debuggerPath", "debuggerArgs", "sourceDirs", "tools"},
        ),
    )
    debugger = Debugger(dap_factory, launch_arguments, config.tools)

    def ensure_file_path(str_path: str) -> Path | None:
        path = Path(str_path)
        if not path.is_file():
            if path.is_absolute():
                return None
            for potential_dir in config.sourceDirs:
                potential_path = Path(potential_dir) / path
                if potential_path.is_file():
                    return potential_path
            return None
        return path

    async def launch():
        return (await debugger.launch()).render()

    async def set_breakpoint(path: str, line: int, condition: Optional[str] = None):
        file_path = ensure_file_path(path)
        if file_path is None:
            return FunctionCallError(message=f"File ({path}) not found").render()
        response = await debugger.set_breakpoint(file_path, line, condition)
        return try_render(response)

    async def remove_breakpoint(path: str, line: int):
        file_path = ensure_file_path(path)
        if file_path is None:
            return FunctionCallError(message=f"File ({path}) not found").render()
        response = await debugger.remove_breakpoint(file_path, line)
        return try_render(response)

    async def view_file_around_line(line: int, path: Optional[str] = None):
        if path:
            file_path = ensure_file_path(path)
            if file_path is None:
                return FunctionCallError(message=f"File ({path}) not found").render()
        else:
            file_path = None
        response = await debugger.view_file_around_line(file_path, line)
        return try_render(response)

    async def remove_all_breakpoints():
        response = await debugger.remove_all_breakpoints()
        if isinstance(response, FunctionCallError):
            return response.render()
        return "All breakpoints removed"

    async def list_all_breakpoints():
        response = await debugger.list_all_breakpoints()
        if isinstance(response, FunctionCallError):
            return response.render()

        def render_file(file: str, breakpoints: list[SourceBreakpoint]) -> str:
            return render_xml(
                "file",
                [
                    render_xml("breakpoint", None, **sb.model_dump())
                    for sb in breakpoints
                ],
                path=file,
            )

        return render_xml(
            "breakpoints",
            [
                render_file(str(file), breakpoints)
                for file, breakpoints in response.items()
            ],
        )

    async def continue_execution():
        response = await debugger.continue_execution()
        return try_render(response)

    async def step_in():
        response = await debugger.step_in()
        return try_render(response)

    async def step_out():
        response = await debugger.step_out()
        return try_render(response)

    async def next():
        response = await debugger.next()
        return try_render(response)

    async def evaluate(expression: str):
        response = await debugger.evaluate(expression)
        return try_render(response)

    async def change_frame(frameId: int):
        response = await debugger.change_frame(frameId)
        return response.render()

    async def terminate():
        response = await debugger.terminate()
        if isinstance(response, FunctionCallError):
            return response.render()
        return response

    @app.call_tool()
    async def call_tool(
        name: str, arguments: dict
    ) -> list[types.TextContent | types.ImageContent | types.EmbeddedResource]:
        if name == "get_launch_config" and config.tools.getLaunchConfig.enabled:
            config_schema = json.dumps(type(config).model_json_schema())
            config_json = json.dumps(config.model_dump(exclude_none=True))
            return [
                types.TextContent(
                    type="text",
                    text=render_xml(
                        "config",
                        [
                            render_xml("schema", config_schema),
                            render_xml("data", config_json),
                        ],
                    ),
                )
            ]
        if name == "launch" and config.tools.launch.enabled:
            return [types.TextContent(type="text", text=await launch())]
        if name == "set_breakpoint" and config.tools.setBreakpoint.enabled:
            return [
                types.TextContent(
                    type="text",
                    text=await set_breakpoint(
                        arguments["path"], arguments["line"], arguments.get("condition")
                    ),
                )
            ]
        if name == "remove_breakpoint" and config.tools.removeBreakpoint.enabled:
            return [
                types.TextContent(
                    type="text",
                    text=await remove_breakpoint(arguments["path"], arguments["line"]),
                )
            ]
        if name == "view_file_around_line" and config.tools.viewFileAroundLine.enabled:
            return [
                types.TextContent(
                    type="text",
                    text=await view_file_around_line(
                        arguments["line"], arguments.get("path")
                    ),
                )
            ]
        if (
            name == "remove_all_breakpoints"
            and config.tools.removeAllBreakpoints.enabled
        ):
            return [types.TextContent(type="text", text=await remove_all_breakpoints())]
        if name == "list_all_breakpoints" and config.tools.listAllBreakpoints.enabled:
            return [types.TextContent(type="text", text=await list_all_breakpoints())]
        if name == "continue_execution" and config.tools.continueExecution.enabled:
            return [types.TextContent(type="text", text=await continue_execution())]
        if name == "step_in" and config.tools.stepIn.enabled:
            return [types.TextContent(type="text", text=await step_in())]
        if name == "step_out" and config.tools.stepOut.enabled:
            return [types.TextContent(type="text", text=await step_out())]
        if name == "next" and config.tools.next.enabled:
            return [types.TextContent(type="text", text=await next())]
        if name == "evaluate" and config.tools.evaluate.enabled:
            return [
                types.TextContent(
                    type="text", text=await evaluate(arguments["expression"])
                )
            ]
        if name == "change_frame" and config.tools.changeFrame.enabled:
            return [
                types.TextContent(
                    type="text", text=await change_frame(arguments["frameId"])
                )
            ]
        if name == "terminate" and config.tools.terminate.enabled:
            return [types.TextContent(type="text", text=await terminate())]
        raise ValueError(f"Unknown tool: {name}")

    @app.list_tools()
    async def list_tools() -> list[types.Tool]:
        tool_list: list[types.Tool | Literal[False]] = [
            config.tools.getLaunchConfig.enabled
            and types.Tool(
                name="get_launch_config",
                description="Returns the user provided launch configuration along with its detailed schema for a DAP-compatible debugger. The schema includes descriptions for each field.",
                inputSchema={"type": "object", "properties": {}},
            ),
            config.tools.launch.enabled
            and types.Tool(
                name="launch",
                description="Launch the debuggee program. Set breakpoints before launching if necessary.",
                inputSchema={"type": "object", "properties": {}},
            ),
            config.tools.setBreakpoint.enabled
            and types.Tool(
                name="set_breakpoint",
                description="Set a breakpoint at the specified file and line with an optional condition.",
                inputSchema={
                    "type": "object",
                    "required": ["path", "line"],
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "The file path where the breakpoint should be set.",
                        },
                        "line": {
                            "type": "integer",
                            "description": "The line number at which to set the breakpoint.",
                        },
                        "condition": {
                            "type": "string",
                            "description": "Optional condition to trigger the breakpoint.",
                        },
                    },
                },
            ),
            config.tools.removeBreakpoint.enabled
            and types.Tool(
                name="remove_breakpoint",
                description="Remove a breakpoint from the specified file and line.",
                inputSchema={
                    "type": "object",
                    "required": ["path", "line"],
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "The file path from which to remove the breakpoint.",
                        },
                        "line": {
                            "type": "integer",
                            "description": "The line number of the breakpoint to remove.",
                        },
                    },
                },
            ),
            config.tools.listAllBreakpoints.enabled
            and types.Tool(
                name="list_all_breakpoints",
                description="List all breakpoints currently set in the debugger.",
                inputSchema={"type": "object", "properties": {}},
            ),
            config.tools.removeAllBreakpoints.enabled
            and types.Tool(
                name="remove_all_breakpoints",
                description="Remove all breakpoints currently set in the debugger.",
                inputSchema={"type": "object", "properties": {}},
            ),
            config.tools.continueExecution.enabled
            and types.Tool(
                name="continue_execution",
                description="Continue execution in the debugger after hitting a breakpoint.",
                inputSchema={"type": "object", "properties": {}},
            ),
            config.tools.stepIn.enabled
            and types.Tool(
                name="step_in",
                description="Step into the function call in the debugger.",
                inputSchema={"type": "object", "properties": {}},
            ),
            config.tools.stepOut.enabled
            and types.Tool(
                name="step_out",
                description="Step out of the current function in the debugger.",
                inputSchema={"type": "object", "properties": {}},
            ),
            config.tools.next.enabled
            and types.Tool(
                name="next",
                description="Step over to the next line of code in the debugger.",
                inputSchema={"type": "object", "properties": {}},
            ),
            config.tools.evaluate.enabled
            and types.Tool(
                name="evaluate",
                description="Evaluate an expression in the current debugging context.",
                inputSchema={
                    "type": "object",
                    "required": ["expression"],
                    "properties": {
                        "expression": {
                            "type": "string",
                            "description": "The expression to evaluate.",
                        }
                    },
                },
            ),
            config.tools.changeFrame.enabled
            and types.Tool(
                name="change_frame",
                description="Change the current debugging frame to the specified frame ID.",
                inputSchema={
                    "type": "object",
                    "required": ["frameId"],
                    "properties": {
                        "frameId": {
                            "type": "integer",
                            "description": "The ID of the frame to switch to.",
                        }
                    },
                },
            ),
            config.tools.terminate.enabled
            and types.Tool(
                name="terminate",
                description="Terminate the current debugging session.",
                inputSchema={"type": "object", "properties": {}},
            ),
            config.tools.viewFileAroundLine.enabled
            and types.Tool(
                name="view_file_around_line",
                description="Returns the lines of source code and the source code around the specified line. You should ALWAYS prefer this tool if you are reading code. Because it will show the line number, which is crucial for debugging. If 'path' is provided, it opens that file; otherwise, it uses the last specified file. You must provide a file to read the source before launch as no prefilled paths exist.",
                inputSchema={
                    "type": "object",
                    "required": ["line"],
                    "properties": {
                        "line": {
                            "type": "integer",
                            "description": "The line number around which the source code will be displayed.",
                        },
                        "path": {
                            "type": "string",
                            "description": "Optional file path. Provide this to open a specific file; otherwise, the last specified file is used.",
                        },
                    },
                },
            ),
        ]
        return [tool for tool in tool_list if tool]

    if transport == "sse":
        from mcp.server.sse import SseServerTransport
        from starlette.applications import Starlette
        from starlette.routing import Mount, Route

        sse = SseServerTransport("/messages/")
        debugger_inited = False

        async def handle_sse(request):
            nonlocal debugger_inited
            if not debugger_inited:
                await debugger.initialize()
                debugger_inited = True
            async with sse.connect_sse(
                request.scope, request.receive, request._send
            ) as streams:
                await app.run(
                    streams[0], streams[1], app.create_initialization_options()
                )

        starlette_app = Starlette(
            debug=True,
            routes=[
                Route("/sse", endpoint=handle_sse),
                Mount("/messages/", app=sse.handle_post_message),
            ],
        )

        import uvicorn

        uvicorn.run(starlette_app, host="0.0.0.0", port=port)
    else:
        from mcp.server.stdio import stdio_server

        async def arun():
            await debugger.initialize()
            async with stdio_server() as streams:
                await app.run(
                    streams[0], streams[1], app.create_initialization_options()
                )

        anyio.run(arun)

    return 0


if __name__ == "__main__":
    main()



import asyncio.subprocess

from asyncio.subprocess import Process
from typing import Protocol, Optional

from dap_mcp.dap import DAPClient


class DAPFactory(Protocol):
    async def create_instance(self) -> DAPClient: ...

    async def destroy_instance(self, client: DAPClient): ...


class DAPClientSingletonFactory:
    def __init__(self, cmd: str, args: list[str], **kwargs):
        self.cmd = cmd
        self.args = args
        self.kwargs = kwargs
        self.debugger_process: Optional[Process] = None

    async def create_instance(self) -> DAPClient:
        if self.debugger_process is not None:
            raise Exception(
                "DAPClientSingletonFactory can only create one instance of DAPClient"
            )
        adapter = await asyncio.subprocess.create_subprocess_exec(
            self.cmd,
            *self.args,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            **self.kwargs,
        )
        reader = adapter.stdout
        writer = adapter.stdin
        assert reader is not None, "Connection closed"
        assert writer is not None, "Connection closed"
        self.debugger_process = adapter
        return DAPClient(reader, writer)

    async def destroy_instance(self, client: DAPClient):
        if self.debugger_process is None:
            return
        if client.stream_reader != self.debugger_process.stdout:
            raise Exception("Client does not belong to this factory")
        self.debugger_process.terminate()
        self.debugger_process = None



import logging

from dap_types import (
    SetBreakpointsResponse,
    ErrorResponse,
    SetBreakpointsRequest,
    SetBreakpointsArguments,
    Source,
    SourceBreakpoint,
    ContinueRequest,
    ContinueArguments,
    ContinueResponse,
    EvaluateRequest,
    EvaluateArguments,
    EvaluateResponse,
    LaunchRequest,
    LaunchRequestArguments,
    InitializeRequest,
    InitializeRequestArguments,
    InitializeResponse,
    LaunchResponse,
    StoppedEvent,
    TerminatedEvent,
    Event,
    ConfigurationDoneRequest,
    StackTraceRequest,
    StackTraceArguments,
    ThreadsResponse,
    Thread,
    StackFrame,
    ThreadsRequest,
    StackTraceResponse,
    ModuleEvent,
    Module,
    Request,
    Response,
    ScopesResponse,
    ScopesRequest,
    ScopesArguments,
    VariablesRequest,
    VariablesArguments,
    VariablesResponse,
    Scope,
    Variable,
    ExceptionInfoRequest,
    ExceptionInfoArguments,
    ExceptionInfoResponse,
    ExceptionInfoResponseBody,
    SetExceptionBreakpointsRequest,
    SetExceptionBreakpointsArguments,
    SetExceptionBreakpointsResponse,
    StepInRequest,
    StepInArguments,
    StepOutRequest,
    StepOutArguments,
    NextRequest,
    NextArguments,
)
from dataclasses import dataclass
from pathlib import Path
from pydantic import BaseModel
from typing import Literal, Optional, List, Tuple, Callable, Any

from dap_mcp.config import DAPToolsConfig
from dap_mcp.dap import DAPClient
from dap_mcp.factory import DAPFactory
from dap_mcp.render import render_table, render_scope, render_xml, try_dump_base_model


class FunctionCallError(BaseModel):
    message: str

    def render(self) -> str:
        return render_xml("error", self.message)


@dataclass
class EventListView:
    events: List[Event]

    def render(self) -> str:
        return render_xml(
            "events",
            [
                render_xml("event", try_dump_base_model(event.body), type=event.event)
                for event in self.events
            ],
        )


@dataclass
class SourceCodeView:
    source: Path | None
    # the line is in the center of the code view
    source_center_line: int
    # the line has an arrow pointing to it (active line)
    source_active_line: Optional[int]

    def get_source(self) -> Tuple[List[Tuple[int, str]], int]:
        if self.source is None:
            return [], 0
        with open(self.source, "r") as f:
            lines = f.readlines()
        span = 20
        self.source_center_line = min(len(lines), self.source_center_line)
        span_begin = max(1, self.source_center_line - span)
        span_end = min(len(lines), self.source_center_line + span)
        source = list(
            zip(range(span_begin, span_end + 1), lines[span_begin - 1 : span_end])
        )
        if source:
            source[-1] = (source[-1][0], source[-1][1].rstrip())
        return source, len(lines)

    def render(self) -> str:
        source, line_count = self.get_source()
        source_view_content = (
            "!!!source not available!!!"
            if source == []
            else render_table(self.source_active_line, source, line_delimiter="")
        )
        return render_xml(
            "source",
            source_view_content,
            path=str(self.source.resolve()) if self.source else None,
            total_lines=line_count,
        )


@dataclass
class StoppedDebuggerView:
    source: SourceCodeView
    frames: List[StackFrame]
    frame_active_id: int
    threads: List[Thread]
    thread_active_id: int
    # module_map: dict[int, Module]
    variables: List[Tuple[Scope, List[Variable]]]
    events: EventListView
    exception_info: Optional[ExceptionInfoResponseBody]
    render_variables: bool = True
    max_variable_expr_length: Optional[int] = None

    def render(self) -> str:
        return render_xml(
            "debugger_view",
            [self.source.render()]
            + (
                [
                    render_xml(
                        "variables",
                        [
                            render_scope(
                                scope, variables, self.max_variable_expr_length
                            )
                            for scope, variables in self.variables
                        ],
                    )
                ]
                if self.render_variables
                else []
            )
            + [
                render_xml(
                    "frames",
                    render_table(
                        self.frame_active_id,
                        [
                            (
                                frame.id,
                                "\t".join(
                                    [frame.name]
                                    + (
                                        [f"path={frame.source.path}"]
                                        if frame.source
                                        else []
                                    )
                                    + [f"line={frame.line}", f"col={frame.column}"]
                                ),
                            )
                            for frame in self.frames
                        ],
                    ),
                ),
                render_xml(
                    "threads",
                    render_table(
                        self.thread_active_id,
                        [(thread.id, thread.name) for thread in self.threads],
                    ),
                ),
                self.events.render(),
            ]
            + (
                [
                    render_xml(
                        "exception_info",
                        f"{self.exception_info.exceptionId} {self.exception_info.description}",
                    )
                ]
                if self.exception_info
                else []
            ),
        )


DebuggerState = Literal[
    "before_initialization", "before_launch", "launched", "errored", "stopped"
]


class MethodWithAvailableStates:
    available_states: List[DebuggerState]

    async def __call__(self, *arg, **kwargs) -> Any: ...


class Debugger:
    # launched is an internal state that is not exposed to the LLM
    def __init__(
        self,
        factory: DAPFactory,
        launch_arguments: LaunchRequestArguments,
        config_or_none: DAPToolsConfig | None = None,
    ):
        if config_or_none is None:
            self._config = DAPToolsConfig()
        else:
            self._config = config_or_none
        self.state: DebuggerState = "before_initialization"
        self._factory = factory
        self._client: Optional[DAPClient] = None
        self.breakpoints: dict[Path, list[SourceBreakpoint]] = {}
        self.launch_arguments = launch_arguments
        self.launch_request: Optional[LaunchRequest] = None
        self.modules: dict[int, Module] = {}
        self.variables: dict[int, Variable] = {}
        self.events: list[Event] = []
        self.frames: dict[int, StackFrame] = {}
        self.active_frame_id: Optional[int] = None
        self.active_thread_id: Optional[int] = None
        self.alternative_center_line_to_view: Optional[int] = None
        self.alternative_file_to_view: Optional[Path] = None

    def _get_available_actions(self) -> List[str]:
        func_names: List[str] = []
        for func_name in dir(self):
            func = getattr(self, func_name)
            if hasattr(func, "available_states") and isinstance(
                func.available_states, list
            ):
                if self.state in func.available_states:
                    func_names.append(func_name)
        return func_names

    @staticmethod
    def available_states(operation: str, states: List[DebuggerState]):
        def decorator(function: Callable[..., Any]):
            async def wrapper(self, *args, **kwargs):
                if self.state not in states:
                    return FunctionCallError(
                        message=f"Cannot {operation} in {self.state} state\nAvailable actions for {self.state}: {self._get_available_actions()}."
                    )
                return await function(self, *args, **kwargs)

            wrapper: MethodWithAvailableStates  # type: ignore[no-redef]
            wrapper.available_states = states  # type: ignore
            return wrapper

        return decorator

    async def _handle_events(self, events: list[Event]):
        for event in events:
            if isinstance(event, ModuleEvent):
                if event.body.reason != "remove":
                    self.modules[event.body.module.id] = event.body.module
                else:
                    self.modules.pop(event.body.module.id, None)
            if isinstance(event, StoppedEvent):
                self.active_thread_id = event.body.threadId
                logging.debug(f"Debugger state changed from {self.state} to stopped")
                self.state = "stopped"
                if event.body.reason == "exception":
                    self.exception_raised = True
            elif isinstance(event, TerminatedEvent):
                logging.debug(
                    f"Debugger state changed from {self.state} to before_initialization"
                )
                self.state = "before_initialization"
                await self._terminate()
        self.events += events

    def _pop_events(self) -> list[Event]:
        events = self.events
        self.events = []
        return events

    async def _send_request(self, request: Request):
        assert self._client is not None, "Client is not initialized"
        return await self._client.send_request(request)

    async def _wait_for_request(self, request: Request) -> Response | ErrorResponse:
        assert self._client is not None, "Client is not initialized"
        response, events = await self._client.wait_for_request(request)
        await self._handle_events(events)
        if isinstance(response, StackTraceResponse):
            self.active_frame_id = response.body.stackFrames[0].id
        elif isinstance(response, StackTraceResponse):
            for frame in response.body.stackFrames:
                self.frames[frame.id] = frame
        elif isinstance(response, LaunchResponse) or isinstance(
            response, ContinueResponse
        ):
            self.active_frame_id = None
            self.active_thread_id = None
            self.frames = {}
            logging.debug(f"Debugger state changed from {self.state} to launched")
            self.state = "launched"
        return response

    async def _wait_for_event_types(self, event_types: set[str]):
        assert self._client is not None, "Client is not initialized"
        events = await self._client.wait_for_event_types(event_types)
        await self._handle_events(events)

    async def initialize(self) -> InitializeResponse:
        self._client = await self._factory.create_instance()
        request = InitializeRequest(
            seq=0,
            type="request",
            command="initialize",
            arguments=InitializeRequestArguments(
                clientID="dap_mcp",
                clientName="dap_mcp",
                adapterID="debugpy",
                locale="en",
            ),
        )
        await self._send_request(request)
        response: ErrorResponse | InitializeResponse = await self._wait_for_request(
            request
        )
        if isinstance(response, ErrorResponse):
            logging.fatal(f"Failed to initialize debugger: {response.message}")
            raise RuntimeError("Failed to initialize debugger")
        logging.debug(f"Debugger state changed from {self.state} to before_launch")
        self.state = "before_launch"
        self.launch_request = await self._send_launch_request()
        return response

    async def _update_breakpoints(
        self, path: Path
    ) -> SetBreakpointsResponse | ErrorResponse:
        breakpoints = self.breakpoints.get(path, [])
        request = SetBreakpointsRequest(
            seq=0,
            type="request",
            command="setBreakpoints",
            arguments=SetBreakpointsArguments(
                source=Source(
                    name=path.name,
                    path=str(path.resolve()),
                ),
                breakpoints=breakpoints,
            ),
        )
        await self._send_request(request)
        response: ErrorResponse | SetBreakpointsResponse = await self._wait_for_request(
            request
        )
        return response

    async def _get_threads(self) -> ThreadsResponse:
        if self.state != "stopped":
            raise RuntimeError(f"Cannot get threads in {self.state} state")
        request = ThreadsRequest(seq=0, type="request", command="threads")
        await self._send_request(request)
        response: ThreadsResponse | ErrorResponse = await self._wait_for_request(
            request
        )
        if isinstance(response, ErrorResponse):
            raise RuntimeError("Error handing for threads is not implemented")
        return response

    async def _get_stack_trace(self, threadId: int) -> StackTraceResponse:
        if self.state != "stopped":
            raise RuntimeError(f"Cannot get stack trace in {self.state} state")
        request = StackTraceRequest(
            seq=0,
            type="request",
            command="stackTrace",
            arguments=StackTraceArguments(
                threadId=threadId,
                startFrame=0,
            ),
        )
        await self._send_request(request)
        response: ErrorResponse | StackTraceResponse = await self._wait_for_request(
            request
        )
        if isinstance(response, ErrorResponse):
            raise RuntimeError("Error handing for stack trace is not implemented")
        return response

    async def _get_scopes(self, frameId: int) -> ScopesResponse:
        request = ScopesRequest(
            seq=0,
            type="request",
            command="scopes",
            arguments=ScopesArguments(frameId=frameId),
        )
        await self._send_request(request)
        response: ErrorResponse | ScopesResponse = await self._wait_for_request(request)
        if isinstance(response, ErrorResponse):
            raise RuntimeError("Error handing for scopes is not implemented")
        return response

    async def _get_variables(self, variablesReference: int) -> VariablesResponse:
        request = VariablesRequest(
            seq=0,
            type="request",
            command="variables",
            arguments=VariablesArguments(variablesReference=variablesReference),
        )
        await self._send_request(request)
        response: ErrorResponse | VariablesResponse = await self._wait_for_request(
            request
        )
        if isinstance(response, ErrorResponse):
            raise RuntimeError("Error handing for variables is not implemented")
        for variable in response.body.variables:
            if variable.variablesReference != 0:
                self.variables[variable.variablesReference] = variable
        return response

    async def _get_exception_info(self) -> ExceptionInfoResponse:
        request = ExceptionInfoRequest(
            seq=0,
            type="request",
            command="exceptionInfo",
            arguments=ExceptionInfoArguments(threadId=self.active_thread_id),
        )
        await self._send_request(request)
        response: ErrorResponse | ExceptionInfoResponse = await self._wait_for_request(
            request
        )
        if isinstance(response, ErrorResponse):
            raise RuntimeError("Error handing for exception info is not implemented")
        return response

    async def _get_stopped_debugger_view(self):
        threads = await self._get_threads()
        assert self.active_thread_id is not None, (
            "Active thread is not set, this should be set via Events"
        )
        stack_trace = await self._get_stack_trace(self.active_thread_id)
        assert self.active_frame_id is not None, (
            "Active frame is not set, this should be set via Events"
        )
        if self._config.debuggerView.showVariables:
            scopes = await self._get_scopes(self.active_frame_id)
            variables = [
                (
                    scope,
                    (
                        await self._get_variables(scope.variablesReference)
                    ).body.variables,
                )
                for scope in scopes.body.scopes
            ]
        else:
            variables = []
        if self.state == "errored":
            exception_info = (await self._get_exception_info()).body
        else:
            exception_info = None
        current_frame = next(
            f for f in stack_trace.body.stackFrames if f.id == self.active_frame_id
        )
        return StoppedDebuggerView(
            source=SourceCodeView(
                source=self.alternative_file_to_view or Path(current_frame.source.path),
                source_center_line=self.alternative_center_line_to_view
                or current_frame.line,
                source_active_line=current_frame.line,
            ),
            frames=stack_trace.body.stackFrames,
            frame_active_id=self.active_frame_id,
            threads=threads.body.threads,
            thread_active_id=self.active_thread_id,
            variables=variables,
            events=EventListView(events=self._pop_events()),
            exception_info=exception_info,
            render_variables=self._config.debuggerView.showVariables,
        )

    @available_states("set breakpoint", ["before_launch", "stopped", "errored"])
    async def set_breakpoint(
        self, path: Path, line: int, condition: Optional[str] = None
    ) -> FunctionCallError | SetBreakpointsResponse | ErrorResponse:
        breakpoints = self.breakpoints.get(path, [])
        for b in breakpoints:
            if b.condition == condition and b.line == line:
                return FunctionCallError(message="Breakpoint already exists")
        breakpoints.append(SourceBreakpoint(line=line, condition=condition))
        self.breakpoints[path] = breakpoints
        return await self._update_breakpoints(path)

    @available_states("remove breakpoint", ["before_launch", "stopped", "errored"])
    async def remove_breakpoint(
        self, path: Path, line: int
    ) -> FunctionCallError | SetBreakpointsResponse | ErrorResponse:
        breakpoints = self.breakpoints.get(path, [])
        breakpoints = [b for b in breakpoints if b.line != line]
        self.breakpoints[path] = breakpoints
        return await self._update_breakpoints(path)

    @available_states("list all breakpoints", ["before_launch", "stopped", "errored"])
    async def list_all_breakpoints(
        self,
    ) -> dict[Path, list[SourceBreakpoint]] | FunctionCallError:
        return self.breakpoints

    @available_states("remove all breakpoints", ["before_launch", "stopped", "errored"])
    async def remove_all_breakpoints(
        self,
    ) -> None | FunctionCallError:
        for path in self.breakpoints:
            await self._update_breakpoints(path)
        self.breakpoints = {}
        return None

    @available_states("continue execution", ["stopped"])
    async def continue_execution(
        self,
    ) -> FunctionCallError | StoppedDebuggerView | EventListView:
        request = ContinueRequest(
            seq=0,
            type="request",
            command="continue",
            arguments=ContinueArguments(threadId=1),
        )
        await self._send_request(request)
        response = await self._wait_for_request(request)
        if isinstance(response, ErrorResponse):
            raise RuntimeError("Error handing for continue is not implemented")
        self.alternative_center_line_to_view = None
        self.alternative_file_to_view = None
        await self._wait_for_event_types({"stopped", "terminated"})
        if self.state != "stopped":
            return EventListView(events=self._pop_events())
        return await self._get_stopped_debugger_view()

    @available_states("step in", ["stopped"])
    async def step_in(
        self,
    ) -> FunctionCallError | StoppedDebuggerView | EventListView:
        request = StepInRequest(
            seq=0,
            type="request",
            command="stepIn",
            arguments=StepInArguments(threadId=self.active_thread_id),
        )
        await self._send_request(request)
        response = await self._wait_for_request(request)
        if isinstance(response, ErrorResponse):
            raise RuntimeError("Error handing for step in is not implemented")
        self.alternative_center_line_to_view = None
        self.alternative_file_to_view = None
        await self._wait_for_event_types({"stopped", "terminated"})
        if self.state != "stopped":
            return EventListView(events=self._pop_events())
        return await self._get_stopped_debugger_view()

    @available_states("step out", ["stopped"])
    async def step_out(
        self,
    ) -> FunctionCallError | StoppedDebuggerView | EventListView:
        request = StepOutRequest(
            seq=0,
            type="request",
            command="stepOut",
            arguments=StepOutArguments(threadId=self.active_thread_id),
        )
        await self._send_request(request)
        response = await self._wait_for_request(request)
        if isinstance(response, ErrorResponse):
            raise RuntimeError("Error handing for step out is not implemented")
        self.alternative_center_line_to_view = None
        self.alternative_file_to_view = None
        await self._wait_for_event_types({"stopped", "terminated"})
        if self.state != "stopped":
            return EventListView(events=self._pop_events())
        return await self._get_stopped_debugger_view()

    @available_states("next", ["stopped"])
    async def next(
        self,
    ) -> FunctionCallError | StoppedDebuggerView | EventListView:
        request = NextRequest(
            seq=0,
            type="request",
            command="next",
            arguments=NextArguments(threadId=self.active_thread_id),
        )
        await self._send_request(request)
        response = await self._wait_for_request(request)
        if isinstance(response, ErrorResponse):
            raise RuntimeError("Error handing for next is not implemented")
        self.alternative_center_line_to_view = None
        self.alternative_file_to_view = None
        await self._wait_for_event_types({"stopped", "terminated"})
        if self.state != "stopped":
            return EventListView(events=self._pop_events())
        return await self._get_stopped_debugger_view()

    @available_states("evaluate an expression", ["stopped"])
    async def evaluate(
        self, expression: str
    ) -> FunctionCallError | EvaluateResponse | ErrorResponse:
        request = EvaluateRequest(
            seq=0,
            type="request",
            command="evaluate",
            arguments=EvaluateArguments(
                expression=expression, frameId=self.active_frame_id
            ),
        )
        await self._send_request(request)
        response: ErrorResponse | EvaluateResponse = await self._wait_for_request(
            request
        )
        return response

    async def _send_launch_request(self) -> LaunchRequest:
        if self.state != "before_launch":
            raise RuntimeError(f"Cannot launch in {self.state} state")
        request = LaunchRequest(
            seq=0, type="request", command="launch", arguments=self.launch_arguments
        )
        await self._send_request(request)
        return request

    @available_states("launch", ["before_launch"])
    async def launch(self) -> FunctionCallError | StoppedDebuggerView | EventListView:
        for path in self.breakpoints:
            await self._update_breakpoints(path)
        set_exception_breakpoints_request = SetExceptionBreakpointsRequest(
            seq=0,
            type="request",
            command="setExceptionBreakpoints",
            arguments=SetExceptionBreakpointsArguments(filters=["uncaught"]),
        )
        await self._send_request(set_exception_breakpoints_request)
        set_exception_breakpoints_response: (
            SetExceptionBreakpointsResponse | ErrorResponse
        ) = await self._wait_for_request(set_exception_breakpoints_request)
        if isinstance(set_exception_breakpoints_response, ErrorResponse):
            raise RuntimeError(
                "Error handing for set exception breakpoints is not implemented"
            )
        configure_done_request = ConfigurationDoneRequest(
            seq=0, type="request", command="configurationDone"
        )
        await self._send_request(configure_done_request)
        configure_done_response: (
            ErrorResponse | ConfigurationDoneRequest
        ) = await self._wait_for_request(configure_done_request)
        if isinstance(configure_done_response, ErrorResponse):
            raise RuntimeError(
                "Error handing for configuration done is not implemented"
            )
        launch_response: ErrorResponse | LaunchResponse = await self._wait_for_request(
            self.launch_request
        )
        if isinstance(launch_response, ErrorResponse):
            raise RuntimeError("Error handing for launch is not implemented")
        self.alternative_center_line_to_view = None
        self.alternative_file_to_view = None
        await self._wait_for_event_types({"stopped", "terminated"})
        if self.state != "stopped":
            return EventListView(events=self._pop_events())
        return await self._get_stopped_debugger_view()

    @available_states("change frame", ["stopped", "errored"])
    async def change_frame(
        self, frame_id: int
    ) -> FunctionCallError | StoppedDebuggerView:
        if frame_id not in self.frames:
            return FunctionCallError(
                message=f"Frame {frame_id} not found, available frames: {list(self.frames.keys())}"
            )
        self.active_frame_id = frame_id
        self.alternative_center_line_to_view = None
        self.alternative_file_to_view = None
        return await self._get_stopped_debugger_view()

    @available_states(
        "view file around line",
        ["before_initialization", "before_launch", "launched", "errored", "stopped"],
    )
    async def view_file_around_line(
        self, file: Optional[Path], line: int
    ) -> SourceCodeView | FunctionCallError:
        if (
            file is None
            and self.alternative_file_to_view is None
            and self.state in ["before_initialization", "before_launch", "launched"]
        ):
            return FunctionCallError(
                message=f"You must provide a file path to view in {self.state} state"
            )
        self.alternative_center_line_to_view = line
        if file:
            self.alternative_file_to_view = file
        current_active_line = (
            self.frames[self.active_frame_id].line if self.active_frame_id else None
        )
        source = SourceCodeView(
            source=self.alternative_file_to_view,
            source_center_line=line,
            source_active_line=current_active_line,
        )
        return source

    @available_states("terminate", ["stopped", "errored"])
    async def terminate(self) -> str | FunctionCallError:
        return await self._terminate()

    async def _terminate(self) -> str:
        logging.debug(
            f"terminating debugger... self._client is not None: {self._client is not None}"
        )
        if self._client:
            await self._factory.destroy_instance(self._client)
            self._client = None
        # Keep breakpoints
        # self.breakpoints = {}
        self.launch_request = None
        self.modules = {}
        self.variables = {}
        self.events = []
        self.frames = {}
        self.active_frame_id = None
        self.active_thread_id = None
        self.alternative_center_line_to_view = None
        self.alternative_file_to_view = None
        logging.debug(
            f"Debugger initialized, state before initialization: {self.state}"
        )
        await self.initialize()
        logging.debug(f"Debugger initialized, state after initialization: {self.state}")
        return "Debugger terminated"



import json
from dap_types import Variable, Scope, ErrorResponse, Response
from pydantic import BaseModel
from typing import Optional, Tuple, Any, Callable, Protocol


class RenderableContent(Protocol):
    def render(self) -> str: ...


def is_plain_object(obj: Any) -> bool:
    if obj is None:
        return True
    if isinstance(obj, (str, int, float, bool)):
        return True
    if isinstance(obj, list):
        return all(is_plain_object(item) for item in obj)
    if isinstance(obj, dict):
        return all(isinstance(k, str) and is_plain_object(v) for k, v in obj.items())
    return False


def try_dump_base_model(
    model: Any, fallback: Callable[[Any], str] = str
) -> str | list[str] | None:
    if model is None:
        return None
    if isinstance(model, BaseModel):
        return json.dumps(model.model_dump(exclude_none=True))
    if isinstance(model, list):
        if not is_plain_object(model):
            raise ValueError("Cannot dump a list of non-plain objects")
        return json.dumps(model)
    if isinstance(model, dict):
        if not is_plain_object(model):
            raise ValueError("Cannot dump a dict of non-plain objects")
        return json.dumps(
            {
                k: try_dump_base_model(v, fallback)
                for k, v in filter(lambda kv: kv[1] is not None, model.items())
            }
        )
    return fallback(model)


def render_xml(tag: str, content: str | list[str] | None, **attrs) -> str:
    filtered_attr_tuples = [(k, v) for k, v in attrs.items() if v is not None]
    attr_str = " ".join([f'{k}="{v}"' for k, v in filtered_attr_tuples])
    tag_with_attr_str = f"{tag} {attr_str}" if attr_str != "" else tag
    if content:
        if isinstance(content, list):
            content = "\n".join(content)
        return f"<{tag_with_attr_str}>{content}</{tag}>"
    return f"<{tag_with_attr_str}/>"


def render_variable(variable: Variable, max_variable_expr_length: Optional[int]) -> str:
    value = variable.value
    if max_variable_expr_length is not None and len(value) > max_variable_expr_length:
        value = value[:max_variable_expr_length] + "...(truncated)"
    return render_xml(
        "variable", value, **variable.model_dump(exclude_none=True, exclude={"value"})
    )


def render_scope(
    scope: Scope, variables: list[Variable], max_variable_expr_length: Optional[int]
) -> str:
    return render_xml(
        "scope",
        [render_variable(variable, max_variable_expr_length) for variable in variables],
        name=scope.name,
        line=scope.line,
        column=scope.column,
    )


def render_table(
    active_id: Optional[int], lines: list[Tuple[int, str]], line_delimiter: str = "\n"
) -> str:
    # print lines with active line marked
    # active_id is the index of the active line in lines.
    # lines: [(line_number, line_content)].
    # If active_id equals to the index of the line in lines, that line is active.
    # It will print in the following format:
    # {line_number} -> {line_content}
    # {line_number}    {line_content} (otherwise)
    if len(lines) == 0:
        return ""
    max_line_number = max([line_number for line_number, _ in lines])
    max_line_number_length = len(str(max_line_number))
    formatted_lines = []
    for line_number, line_content in lines:
        if line_number == active_id:
            formatted_lines.append(
                f"{line_number:>{max_line_number_length}} -> {line_content}"
            )
        else:
            formatted_lines.append(
                f"{line_number:>{max_line_number_length}}    {line_content}"
            )
    return "\n" + line_delimiter.join(formatted_lines) + "\n"


def render_response(r: Response) -> str:
    if isinstance(r, ErrorResponse):
        return render_xml(
            "error", try_dump_base_model(r.body), command=r.command, message=r.message
        )
    return render_xml(
        "response", try_dump_base_model(r.body), command=r.command, message=r.message
    )


def try_render(r: Response | RenderableContent) -> str:
    if isinstance(r, Response):
        return render_response(r)
    else:
        return r.render()



import json
import pydantic_core
from asyncio import StreamReader, StreamWriter
from dap_types import Request, Response, Event, DiscriminatedProtocolMessage
from pydantic import TypeAdapter, ValidationError, Field
from typing import Tuple, Union, Annotated
import logging

logger = logging.getLogger(__name__)


class DAPClient:
    def __init__(self, stream_reader: StreamReader, stream_writer: StreamWriter):
        self.seq = 1
        self.request_sent: dict[int, Request] = {}
        self.response_received: dict[int, Response] = {}
        self.stream_reader = stream_reader
        self.stream_writer = stream_writer
        self.protocol_message_adapter = TypeAdapter(DiscriminatedProtocolMessage)
        self.lossy_protocol_message_adapter: TypeAdapter[
            Annotated[Union[Request, Event, Response], Field(discriminator="type")]
        ] = TypeAdapter(
            Annotated[Union[Request, Event, Response], Field(discriminator="type")]
        )

    async def send_request(self, request: Request):
        request.seq = self.seq
        self.request_sent[self.seq] = request
        self.seq += 1

        encoded = json.dumps(
            pydantic_core.to_jsonable_python(request, exclude_none=True)
        ).encode("utf-8")
        headers = f"Content-Length: {len(encoded)}\r\n\r\n".encode("utf-8")
        message = headers + encoded

        self.stream_writer.write(message)
        await self.stream_writer.drain()
        logger.debug(f"client -> server: {encoded.decode('utf-8')}")

    async def receive(self) -> Request | Response | Event:
        while True:
            content_length_bytes = await self.stream_reader.readuntil(b"\r\n\r\n")
            assert content_length_bytes is not None, "Connection closed"
            content_length = int(content_length_bytes.decode("utf-8").split(": ")[1])
            message = await self.stream_reader.readexactly(content_length)
            logger.debug(f"server -> client: {message.decode('utf-8')}")
            message = json.loads(message)
            try:
                protocol_message: Request | Response | Event = (
                    self.protocol_message_adapter.validate_python(message)
                )
                return protocol_message
            except ValidationError:
                try:
                    lossy_protocol_message: Request | Response | Event = (
                        self.lossy_protocol_message_adapter.validate_python(message)
                    )
                    return lossy_protocol_message
                except ValidationError:
                    logger.warning(f"Invalid message: {message.decode('utf-8')}")

    @staticmethod
    def _try_discriminate_response(request: Request, response: Response) -> Response:
        if hasattr(request, "discriminate_response"):
            return request.discriminate_response(response)
        return response

    async def wait_for_request(self, request: Request) -> Tuple[Response, list[Event]]:
        if request.seq in self.response_received:
            response = self.response_received.pop(request.seq)
            return self._try_discriminate_response(request, response), []

        events: list[Event] = []
        while True:
            message = await self.receive()
            if isinstance(message, Response):
                if message.request_seq == request.seq:
                    return self._try_discriminate_response(request, message), events
                else:
                    self.response_received[message.request_seq] = message
            elif isinstance(message, Event):
                events.append(message)
            else:
                logger.warning(f"Unexpected message: {message}")

    async def wait_for_event_types(self, event_types: set[str]) -> list[Event]:
        events = []
        while True:
            message = await self.receive()
            if isinstance(message, Event):
                events.append(message)
                if message.event in event_types:
                    return events
            elif isinstance(message, Response):
                self.response_received[message.seq] = message
            else:
                logger.warning(f"Unexpected message: {message}")



from pydantic import BaseModel, Field
from typing import Literal, Optional, List, Dict, Union, Annotated


class DebuggerViewConfig(BaseModel):
    showVariables: bool = Field(
        default=True, description="Show variables in the debugger view."
    )
    maxVariableExprLength: Optional[int] = Field(
        default=None, description="Maximum length of variable expressions."
    )


class BaseToolConfig(BaseModel):
    enabled: bool = Field(default=True, description="Enable or disable the tool.")


class GetLaunchConfigConfig(BaseToolConfig):
    pass


class LaunchConfig(BaseToolConfig):
    pass


class SetBreakpointConfig(BaseToolConfig):
    pass


class RemoveBreakpointConfig(BaseToolConfig):
    pass


class ViewFileAroundLineConfig(BaseToolConfig):
    pass


class ListAllBreakpointsConfig(BaseToolConfig):
    pass


class RemoveAllBreakpointsConfig(BaseToolConfig):
    pass


class ContinueExecutionConfig(BaseToolConfig):
    pass


class StepInConfig(BaseToolConfig):
    pass


class StepOutConfig(BaseToolConfig):
    pass


class NextConfig(BaseToolConfig):
    pass


class EvaluateConfig(BaseToolConfig):
    pass


class ChangeFrameConfig(BaseToolConfig):
    pass


class TerminateConfig(BaseToolConfig):
    pass


class DAPToolsConfig(BaseModel):
    debuggerView: DebuggerViewConfig = Field(
        default_factory=DebuggerViewConfig, description="Debugger view configuration."
    )
    getLaunchConfig: GetLaunchConfigConfig = Field(
        default_factory=GetLaunchConfigConfig
    )
    launch: LaunchConfig = Field(default_factory=LaunchConfig)
    setBreakpoint: SetBreakpointConfig = Field(default_factory=SetBreakpointConfig)
    removeBreakpoint: RemoveBreakpointConfig = Field(
        default_factory=RemoveBreakpointConfig
    )
    viewFileAroundLine: ViewFileAroundLineConfig = Field(
        default_factory=ViewFileAroundLineConfig
    )
    listAllBreakpoints: ListAllBreakpointsConfig = Field(
        default_factory=ListAllBreakpointsConfig
    )
    removeAllBreakpoints: RemoveAllBreakpointsConfig = Field(
        default_factory=RemoveAllBreakpointsConfig
    )
    continueExecution: ContinueExecutionConfig = Field(
        default_factory=ContinueExecutionConfig
    )
    stepIn: StepInConfig = Field(default_factory=StepInConfig)
    stepOut: StepOutConfig = Field(default_factory=StepOutConfig)
    next: NextConfig = Field(default_factory=NextConfig)
    evaluate: EvaluateConfig = Field(default_factory=EvaluateConfig)
    changeFrame: ChangeFrameConfig = Field(default_factory=ChangeFrameConfig)
    terminate: TerminateConfig = Field(default_factory=TerminateConfig)


class DAPConfig(BaseModel):
    type: str
    debuggerPath: str = Field(..., description="Path to the debugger executable.")
    debuggerArgs: List[str] = Field(
        [], description="List of arguments to pass to the debugger executable."
    )
    sourceDirs: List[str] = Field(
        default=[],
        description="List of source directories. Will be used to resolve source paths for set_breakpoint, remove_breakpoint, and view_file_around_line when given paths are relative.",
    )
    tools: DAPToolsConfig = Field(
        default_factory=DAPToolsConfig, description="Tool specific configuration."
    )


class DebugPy(DAPConfig):
    type: Literal["debugpy"]

    # Code Execution Settings
    module: Optional[str] = Field(
        None, description="Name of the module to be debugged."
    )
    program: Optional[str] = Field(None, description="Absolute path to the program.")
    code: Optional[str] = Field(
        None,
        description='Code to execute in string form. Example: "import debugpy;print(debugpy.__version__)"',
    )
    python: List[str] = Field(
        ...,
        description='Path python executable and interpreter arguments. Example: ["/usr/bin/python", "-E"].',
    )
    args: Optional[List[str]] = Field(
        None,
        description='Command line arguments passed to the program. Example: ["--arg1", "-arg2", "val", ...].',
    )
    console: Optional[
        Literal["internalConsole", "integratedTerminal", "externalTerminal"]
    ] = Field(
        "internalConsole",
        description='Sets where to launch the debug target. Supported values: ["internalConsole", "integratedTerminal", "externalTerminal"]. Default is "internalConsole".',
    )
    cwd: Optional[str] = Field(
        None,
        description="Absolute path to the working directory of the program being debugged.",
    )
    env: Optional[Dict[str, str]] = Field(
        None, description="Environment variables defined as a key value pair."
    )

    # Debugger Settings
    django: bool = Field(
        False, description="When true enables Django templates. Default is false."
    )
    gevent: bool = Field(
        False,
        description="When true enables debugging of gevent monkey-patched code. Default is false.",
    )
    jinja: bool = Field(
        False,
        description="When true enables Jinja2 template debugging (e.g. Flask). Default is false.",
    )
    justMyCode: bool = Field(
        True,
        description="When true debug only user-written code. To debug standard library or anything outside of 'cwd' use false. Default is true.",
    )
    logToFile: bool = Field(
        False,
        description="When true enables logging of debugger events to a log file(s). Default is false.",
    )
    pathMappings: Optional[List[Dict[str, str]]] = Field(
        None,
        description="Map of local and remote paths. Example: [{'localRoot': 'local path', 'remoteRoot': 'remote path'}].",
    )
    pyramid: bool = Field(
        False,
        description="When true enables debugging Pyramid applications. Default is false.",
    )
    redirectOutput: bool = Field(
        False,
        description="When true redirects output to debug console. Default is false.",
    )
    showReturnValue: bool = Field(
        False,
        description="Shows return value of functions when stepping. The return value is added to the response to Variables Request.",
    )
    stopOnEntry: bool = Field(
        False,
        description="When true debugger stops at first line of user code. When false debugger does not stop until breakpoint, exception or pause.",
    )
    subProcess: bool = Field(
        True,
        description="When true enables debugging multiprocess applications. Default is true.",
    )
    sudo: bool = Field(
        False,
        description="When true runs program under elevated permissions (on Unix). Default is false.",
    )


class LLDB(DAPConfig):
    type: Literal["lldb"]
    name: str = Field(
        ..., description="A configuration name that will be displayed in the IDE."
    )
    request: Literal["launch", "attach"] = Field(
        ...,
        description='Specifies the request type. Must be either "launch" or "attach".',
    )
    program: str = Field(
        ..., description="Path to the executable to launch or attach to."
    )
    # Source mapping and working directory settings
    sourcePath: Optional[str] = Field(
        None,
        description='Specify a source path to remap "./" for resolving breakpoints in binaries with relative source paths.',
    )
    sourceMap: Optional[List[List[str]]] = Field(
        None,
        description="Array of source path re-mappings. Each element is a two-element array: [source, destination].",
    )
    debuggerRoot: Optional[str] = Field(
        None,
        description="Working directory for launching lldb-dap, used to resolve relative paths in debug information.",
    )
    # LLDB command and formatting settings
    commandEscapePrefix: Optional[str] = Field(
        "`",
        description="Escape prefix for executing regular LLDB commands in the Debug Console. Defaults to a backtick.",
    )
    customFrameFormat: Optional[str] = Field(
        None,
        description="Custom format string for generating stack frame descriptions.",
    )
    customThreadFormat: Optional[str] = Field(
        None, description="Custom format string for generating thread descriptions."
    )
    displayExtendedBacktrace: Optional[bool] = Field(
        False, description="Enable language-specific extended backtraces."
    )
    enableAutoVariableSummaries: Optional[bool] = Field(
        False, description="Enable auto-generated summaries for variables."
    )
    enableSyntheticChildDebugging: Optional[bool] = Field(
        False,
        description="When true, also display raw variable contents for synthetic children.",
    )
    # LLDB command sequences
    initCommands: Optional[List[str]] = Field(
        None,
        description="LLDB commands executed upon debugger startup before target creation.",
    )
    preRunCommands: Optional[List[str]] = Field(
        None,
        description="LLDB commands executed just before launching/attaching, after the target has been created.",
    )
    stopCommands: Optional[List[str]] = Field(
        None, description="LLDB commands executed immediately after each stop."
    )
    exitCommands: Optional[List[str]] = Field(
        None, description="LLDB commands executed when the program exits."
    )
    terminateCommands: Optional[List[str]] = Field(
        None, description="LLDB commands executed when the debugging session ends."
    )
    # Launch configuration settings (for "launch" request)
    args: Optional[List[str]] = Field(
        None, description="Command line arguments passed to the program."
    )
    cwd: Optional[str] = Field(
        None, description="Working directory of the program being launched."
    )
    env: Optional[Dict[str, str]] = Field(
        None,
        description="Environment variables to set when launching the program. Format: {'VAR': 'VALUE'}.",
    )
    stopOnEntry: Optional[bool] = Field(
        False, description="Whether to stop the program immediately after launch."
    )
    runInTerminal: Optional[bool] = Field(
        False,
        description="Launch the program in an integrated terminal. Useful for interactive programs.",
    )
    launchCommands: Optional[List[str]] = Field(
        None, description="LLDB commands executed to launch the program."
    )
    # Attach configuration settings (for "attach" request)
    pid: Optional[int] = Field(
        None,
        description="The process ID of the process to attach to. If omitted, lldb-dap will try to resolve the process by name.",
    )
    waitFor: Optional[bool] = Field(
        False, description="Wait for the process to launch before attaching."
    )
    attachCommands: Optional[List[str]] = Field(
        None, description="LLDB commands executed after preRunCommands during attach."
    )


DebuggerSpecificConfig = Annotated[
    Union[DebugPy, LLDB], Field(..., discriminator="type")
]
